import functools
import sys
import traceback

from stacked import Stacked

from .xtraceback import XTraceback


class TracebackCompat(Stacked):
    """
    A context manager that patches the stdlib traceback module

    Functions in the traceback module that exist as a method of this class are
    replaced with equivalents that use XTraceback.

    :cvar NOPRINT: Exception types that we don't print for (includes None)
    :type NOPRINT: tuple

    :ivar defaults: Default options to apply to XTracebacks created by this
                    instance
    :type defaults: dict
    """

    NOPRINT = (None, KeyboardInterrupt)

    def __init__(self, **defaults):
        super(TracebackCompat, self).__init__()
        self.defaults = defaults
        # register patches for methods that wrap traceback functions
        for key in dir(traceback):
            if hasattr(self, key):
                self._register_patch(traceback, key, getattr(self, key))

    def __exit__(self, etype, evalue, tb):
        if etype not in self.NOPRINT:
            self.print_exception(etype, evalue, tb)
        super(TracebackCompat, self).__exit__(etype, evalue, tb)

    def _factory(self, etype, value, tb, limit=None, **options):
        options["limit"] = \
            getattr(sys, "tracebacklimit", None) if limit is None else limit
        _options = self.defaults.copy()
        _options.update(options)
        return XTraceback(etype, value, tb, **_options)

    def _print_factory(self, etype, value, tb, limit=None, file=None,
                       **options):
        # late binding here may cause problems where there is no sys i.e. on
        # google app engine but it is required for cases where sys.stderr is
        # rebound i.e. under nose
        if file is None and hasattr(sys, "stderr"):
            file = sys.stderr
        options["stream"] = file
        return self._factory(etype, value, tb, limit, **options)

    @functools.wraps(traceback.format_tb)
    def format_tb(self, tb, limit=None, **options):
        xtb = self._factory(None, None, tb, limit, **options)
        return xtb.format_tb()

    @functools.wraps(traceback.format_exception_only)
    def format_exception_only(self, etype, value, **options):
        xtb = self._factory(etype, value, None, **options)
        return xtb.format_exception_only()

    @functools.wraps(traceback.format_exception)
    def format_exception(self, etype, value, tb, limit=None, **options):
        xtb = self._factory(etype, value, tb, limit, **options)
        return xtb.format_exception()

    @functools.wraps(traceback.format_exc)
    def format_exc(self, limit=None, **options):
        options["limit"] = limit
        return "".join(self.format_exception(*sys.exc_info(), **options))

    @functools.wraps(traceback.print_tb)
    def print_tb(self, tb, limit=None, file=None, **options):
        xtb = self._print_factory(None, None, tb, limit, file, **options)
        xtb.print_tb()

    @functools.wraps(traceback.print_exception)
    def print_exception(self, etype, value, tb, limit=None, file=None,
                        **options):
        xtb = self._print_factory(etype, value, tb, limit, file, **options)
        xtb.print_exception()

    @functools.wraps(traceback.print_exc)
    def print_exc(self, limit=None, file=None, **options):
        options["limit"] = limit
        options["file"] = file
        self.print_exception(*sys.exc_info(), **options)



import inspect
import os
import pprint

from .defaults import DEFAULT_WIDTH

#: Filesystem path to the Python standard library
OS_SOURCE = inspect.getsourcefile(os)
STDLIB_PATH = None if OS_SOURCE is None \
    else os.path.dirname(os.path.realpath(OS_SOURCE))

#: A dictionary keyed by type with values as the wrapper for string
#: representation of the type
REFORMAT = {
    dict: ("{", "}"),
    list: ("[", "]"),
    tuple: ("(", ")"),
    set: ("set([", "])"),
    frozenset: ("frozenset([", "])"),
}


def format_filename(options, filename):
    if options.shorten_filenames:
        filename = os.path.realpath(filename)
        if STDLIB_PATH is not None \
                and filename.startswith(STDLIB_PATH):
            filename = filename.replace(STDLIB_PATH, "<stdlib>")
        else:
            # os.path.relpath was introduced in python 2.5
            relative = os.path.relpath(filename)
            if len(relative) < len(filename):
                filename = relative
    # using str on filename to make Jython (default string is unicode)
    # consistent with CPython
    return str(filename)


def format_variable(key, value, indent=4, prefix="", print_width=DEFAULT_WIDTH):

    separator = " = "

    vtype = type(value)
    try:
        pvalue = pprint.pformat(value, indent=0)
    except:
        pvalue = "<unprintable %s object>" % vtype.__name__

    base_size = indent + len(prefix) + len(key) + len(separator)

    if base_size + len(pvalue) > print_width:
        reformat = REFORMAT.get(vtype)
        if reformat is None:
            # truncate long strings - minus 3 for the ellipsis
            pvalue = pvalue[:print_width - base_size - 3] + "..."
        else:
            start, end = reformat
            lines = map(str.strip,
                        pvalue.lstrip(start).rstrip(end).splitlines())
            sub_indent = "\n" + " " * (indent + 4)
            pvalue = "".join((start, sub_indent, sub_indent.join(lines),
                              ",", sub_indent, end))

    return "".join((" " * indent, prefix, key, separator, pvalue))



__version__ = "0.4.0-rc1"

from .loggingcompat import LoggingCompat
from .tracebackcompat import TracebackCompat
from .xtraceback import XTraceback
from .xtracebackoptions import XTracebackOptions

from .pycompat import monkeypatch
monkeypatch()
del monkeypatch

compat = TracebackCompat()



import inspect
import os

from .formatting import format_filename


class ModuleShim(object):

    def __init__(self, options, target):
        self.options = options
        self.target = target

    def __repr__(self):
        package = False
        try:
            filename = inspect.getsourcefile(self.target)
        except TypeError:
            filename = None
        if filename is not None:
            if os.path.basename(filename) == "__init__.py":
                package = True
                filename = os.path.dirname(filename)
            filename = format_filename(self.options, filename)
        if filename is None:
            return repr(self.target)
        return "<%s '%s' from=%r>" % (package and "package" or "module",
                                      self.target.__name__,
                                      filename)



import struct
import traceback
import warnings

try:
    basestring
except NameError:
    # python 3
    basestring = str

try:
    import fcntl
except ImportError:
    fcntl = None
else:
    import termios

try:
    import pygments
except ImportError:
    pygments = None
else:
    from pygments.formatters.terminal import TerminalFormatter
    from pygments.styles import default
    from .lexer import PythonXTracebackLexer

from .defaults import DEFAULT_WIDTH
from .xtracebackoptions import XTracebackOptions
from .xtracebackexc import XTracebackExc


class XTraceback(object):
    """
    An extended traceback formatter

    """

    def __init__(self, etype, value, tb, **options):
        """
        :param etype: The exception type
        :type etype: type
        :param value: The exception instance
        :type value: Exception
        :param tb: The traceback instance
        :type tb: traceback
        :param options: Options for this instance
        :type options: dict
        """

        #: Options for xtraceback
        self.options = XTracebackOptions(**options)

        #: The exception type
        self.etype = etype

        #: The exception value (instance)
        self.value = value

        #: The list of exceptions and tracebacks
        if self.options.chain and self.value is not None \
                and hasattr(traceback, "_iter_chain"):
            # python 3
            values = list(traceback._iter_chain(value, tb))
            # traceback._iter_chain pulls tb from the exception instance and so
            # does not use the tb argument to this method for the last (root)
            # exception; normally this would not matter as the two are the same
            # but it falls over when tb is None so we replace the last tb in
            # the values list with tb from args
            values[-1] = (values[-1][0], tb)
        else:
            values = [(value, tb)]

        #: Used in XTracebackFrame to determine indent
        self.number_padding = 0

        # build list of exceptions
        self.exceptions = []
        for value, tb in values:
            if not isinstance(self.etype, basestring) \
                    and isinstance(value, basestring):
                exc_type = value
                value = None
            else:
                exc_type = etype if value == self.value else type(value)
            exc = XTracebackExc(exc_type, value, tb, self)
            self.exceptions.append(exc)
            self.number_padding = max(exc.number_padding, self.number_padding)

        # placeholders
        self._lexer = None
        self._formatter = None

        # work out print width
        if self.options.print_width is not None:
            self.print_width = self.options.print_width
        elif fcntl is not None and self.tty_stream:
            self.print_width = struct.unpack(
                'HHHH',
                fcntl.ioctl(self.options.stream,
                            termios.TIOCGWINSZ,
                            struct.pack('HHHH', 0, 0, 0, 0)),
            )[1]
        else:
            self.print_width = DEFAULT_WIDTH

    @property
    def tty_stream(self):
        """
        Whether or not our stream is a tty
        """
        return hasattr(self.options.stream, "isatty") \
            and self.options.stream.isatty()

    @property
    def color(self):
        """
        Whether or not color should be output
        """
        return self.tty_stream if self.options.color is None \
            else self.options.color

    def __str__(self):
        return self._str_lines(self._format_exception())

    # { Line formatting

    def _highlight(self, string):
        if pygments is None:
            warnings.warn("highlighting not available - pygments is required")
        else:
            if self._lexer is None:
                self._lexer = PythonXTracebackLexer()
            if self._formatter is None:
                # passing style=default here is the same as passing no
                # arguments - the reason for doing it is that if we don't the
                # style gets imported at runtime which under some restricted
                # environments (appengine) causes a problem - all of the
                # imports must be done before appengine installs its import
                # hook
                self._formatter = TerminalFormatter(style=default)
            try:
                return pygments.highlight(string, self._lexer, self._formatter)
            except KeyboardInterrupt:
                # let the user abort highlighting if problematic
                pass
        return string

    def _str_lines(self, lines):
        exc_str = "".join(lines)
        if self.color:
            exc_str = self._highlight(exc_str)
        return exc_str

    def _format_lines(self, lines):
        # XXX: This is doing the highlight line by line
        return map(self._highlight, lines) if self.color else lines

    def _print_lines(self, lines):
        if self.options.stream is None:
            raise RuntimeError("Cannot print - %r has None stream" % self)
        self.options.stream.write(self._str_lines(lines))

    # { Traceback format - these return lines terminated with "\n"

    def _format_tb(self):
        lines = []
        for exc in self.exceptions:
            lines.extend(exc.format_tb())
        return lines

    def _format_exception_only(self):
        return self.exceptions[-1].format_exception_only()

    def _format_exception(self):
        lines = []
        for exc in self.exceptions:
            lines.extend(exc.format_exception())
        return lines

    # { Interface - this is compatible with the stdlib's traceback module

    def format_tb(self):
        return self._format_lines(self._format_tb())

    def format_exception_only(self):
        return self._format_lines(self._format_exception_only())

    def format_exception(self):
        return self._format_lines(self._format_exception())

    def print_tb(self):
        self._print_lines(self._format_tb())

    def print_exception(self):
        self._print_lines(self._format_exception())

    # }



import inspect
import os
import sys
import textwrap
import types
import warnings

from .formatting import format_filename, format_variable
from .moduleshim import ModuleShim


class XTracebackFrame(object):

    FILTER = ("__builtins__", "__all__", "__doc__", "__file__", "__name__",
              "__package__", "__path__", "__loader__", "__cached__",
              "__initializing__")

    FUNCTION_EXCLUDE = ("GeneratorContextManager.__exit__",)

    GLOBALS_PREFIX = "g:"

    def __init__(self, exc, frame, frame_info, tb_index):

        self.exc = exc
        self.frame = frame
        self.frame_info = frame_info
        self.tb_index = tb_index

        # shortcuts
        self.xtb = exc.xtb
        self.options = self.xtb.options

        # filter variables
        self.locals = self._filter(frame.f_locals)
        self.globals = self._filter(frame.f_globals)

        # filter out globals that are not under the namespace in
        # globals_module_include if the option is not None
        if self.options.globals_module_include is not None:
            for key, value in self.globals.items():
                if isinstance(value, types.ModuleType):
                    module = value.__name__
                elif isinstance(value, types.InstanceType):
                    module = value.__class__.__module__
                else:
                    module = getattr(value, "__module__", None)
                if (module is not None
                        and not module.startswith(
                            self.options.globals_module_include)):
                    del self.globals[key]

        # frame details
        (self.filename, self.lineno, self.function,
         self.code_context, self.index) = frame_info
        self.args, self.varargs, self.varkw = inspect.getargs(frame.f_code)

        # keep track of what we've formatted in this frame
        self._formatted_vars = {}

        # placeholder for formatted frame string
        self._formatted = None

        # if path is a real path then try to shorten it
        if os.path.exists(self.filename):
            self.filename = format_filename(self.options, self.filename)

        # qualify method name with class name
        if self.options.qualify_methods and self.args:
            try:
                cls = self.frame.f_locals[self.args[0]]
            except KeyError:  # pragma: no cover - defensive
                # we're assuming that the first argument is in f_locals but
                # it may not be in some cases so this is a defence, see
                # https://github.com/0compute/xtraceback/issues/3 with further
                # detail at http://www.sqlalchemy.org/trac/ticket/2317 and
                # https://dev.entrouvert.org/issues/765
                pass
            except TypeError:  # pragma: no cover - defensive
                # if self.args[0] is a list it is not hashable -
                # inspect.getargs may return nested lists for args
                pass
            else:
                if not isinstance(cls, type):
                    cls = type(cls)
                if hasattr(cls, self.function):
                    for base in inspect.getmro(cls):
                        if self.function in base.__dict__:
                            self.function = base.__name__ + "." + self.function
                            break

    @property
    def exclude(self):
        return self.locals.get("__xtraceback_skip_frame__", False) \
            or self.function in self.FUNCTION_EXCLUDE

    def _filter(self, fdict):
        try:
            fdict = fdict.copy()
        except NotImplementedError:
            # user data types inheriting dict may not have implemented copy
            pass
        else:
            to_remove = []
            for key, value in fdict.items():
                try:
                    if key in self.FILTER:
                        to_remove.append(key)
                        continue
                except:
                    exc_info = sys.exc_info()
                    # the comparison failed for an unknown reason likely a
                    # custom __cmp__ that makes bad assumptions - swallow
                    try:
                        warnings.warn("Could not filter %r: %r"
                                      % (key, exc_info[1]))
                    except:
                        warnings.warn("Could not filter and can't say why: %s"
                                      % exc_info[1])
                    continue
                else:
                    if isinstance(value, types.ModuleType):
                        value = ModuleShim(self.options, value)
                    elif isinstance(value, dict):
                        value = self._filter(value)
                    fdict[key] = value
            for key in to_remove:
                del fdict[key]
        return fdict

    def _format_variable(self, lines, key, value, indent=4, prefix=""):
        if value is not self._formatted_vars.get(key):
            self._formatted_vars[key] = value
            if self.globals.get(key) is value:
                prefix = self.GLOBALS_PREFIX + prefix
            lines.append(format_variable(key, value, indent, prefix))

    def _format_dict(self, odict, indent=4):
        lines = []
        for key in sorted(odict.keys()):
            self._format_variable(lines, key, odict[key], indent)
        return lines

    def _format_frame(self):

        lines = ['  File "%s", line %d, in %s'
                 % (self.filename, self.lineno, self.function)]

        # push frame args
        if self.options.show_args:
            for arg in self.args:
                if isinstance(arg, list):
                    # TODO: inspect.getargs arg list may contain nested
                    # lists; skip it for now
                    continue
                self._format_variable(lines, arg, self.locals.get(arg))
            if self.varargs:
                self._format_variable(
                    lines, self.varargs, self.locals.get(self.varargs),
                    prefix="*")
            if self.varkw:
                self._format_variable(lines, self.varkw,
                                      self.locals.get(self.varkw), prefix="**")

        # push globals
        if self.options.show_globals:
            lines.extend(self._format_dict(self.globals))

        # push context lines
        if self.code_context is not None:

            lineno = self.lineno - self.index

            dedented = textwrap.dedent("".join(self.code_context))
            for line in dedented.splitlines():

                numbered_line = "    %s" % "%*s %s" \
                    % (self.xtb.number_padding, lineno, line)

                if lineno == self.lineno:

                    if self.options.context > 1:
                        # push the numbered line with a marker
                        dedented_line = numbered_line.lstrip()
                        marker_padding = len(numbered_line) \
                            - len(dedented_line) - 2
                        lines.append("%s> %s" % ("-" * marker_padding,
                                                 dedented_line))
                    else:
                        # push the line only
                        lines.append("    " + line)

                    # push locals below lined up with the start of code
                    if self.options.show_locals:
                        indent = self.xtb.number_padding + len(line) \
                            - len(line.lstrip()) + 5
                        lines.extend(self._format_dict(self.locals, indent))

                else:

                    # push the numbered line
                    lines.append(numbered_line)

                lineno += 1

        elif self.options.show_locals:
            # no context so we are execing
            lines.extend(self._format_dict(self.locals))

        return "\n".join(lines)

    def __str__(self):
        if self._formatted is None:
            self._formatted = self._format_frame()
        return self._formatted



import inspect
import sys

from .formatting import format_filename

from .xtracebackframe import XTracebackFrame


class XTracebackExc(object):

    def __init__(self, etype, value, tb, xtb):
        """
        :param value: The exception instance
        :type value: Exception
        :param tb: The traceback instance
        :type tb: traceback
        :param xtb: the xtraceback instance
        :type xtb: XTraceback
        """

        self.etype = etype

        #: The exception instance
        self.value = value

        #: The XTraceback instance
        self.xtb = xtb

        #: Used in XTracebackFrame to determine indent
        self.number_padding = 0

        #: List of XTracebackFrame
        self.frames = []

        i = 0
        while tb is not None and (self.xtb.options.limit is None
                                  or i < self.xtb.options.limit):
            if i >= self.xtb.options.offset:
                try:
                    frame_info = inspect.getframeinfo(tb,
                                                      self.xtb.options.context)
                except KeyError:  # pragma: no cover - defensive
                    # <stdlib>/inspect.py line 506 - there may be no __main__
                    # XXX: This can't be right - frame_info needs to be defined
                    # in order to construct the XTracebackFrame below.
                    pass
                frame = XTracebackFrame(self, tb.tb_frame, frame_info, i)
                if not frame.exclude:
                    self.frames.append(frame)
                    self.number_padding = max(len(str(frame.lineno)),
                                              self.number_padding)
            tb = tb.tb_next
            i += 1

    # { Traceback format - these return lines that should be joined with ""

    def format_tb(self):
        return ["%s\n" % frame for frame in self.frames]

    def format_exception_only(self):

        lines = []

        if self.value is None:
            value_str = ""
        elif isinstance(self.value, SyntaxError):
            # taken from traceback.format_exception_only
            try:
                msg, (filename, lineno, offset, badline) = self.value.args
            except:
                pass
            else:
                filename = filename and format_filename(self.xtb.options,
                                                        filename) or "<string>"
                filename = filename or "<string>"
                lines.append('  File "%s", line %d\n' % (filename, lineno))
                if badline is not None:
                    lines.append('    %s\n' % badline.strip())
                    if offset is not None:
                        caretspace = badline.rstrip('\n')[:offset].lstrip()
                        # non-space whitespace (likes tabs) must be kept for
                        # alignment
                        caretspace = ((c.isspace() and c or ' ')
                                      for c in caretspace)
                        # only three spaces to account for offset1 == pos 0
                        lines.append('   %s^\n' % ''.join(caretspace))
                    value_str = msg
        else:
            try:
                value_str = str(self.value)
            except Exception:
                try:
                    value_str = unicode(self.value).encode("ascii",
                                                           "backslashreplace")
                except Exception:
                    value_str = "<unprintable %s object>" \
                        % type(self.value).__name__

        # format last line
        if isinstance(self.etype, type):
            stype = self.etype.__name__
            # not using namedtuple to get major version as this is for >= py27
            if sys.version_info[0] == 3:
                # python 3 uses fully-qualified names
                smod = self.etype.__module__
                if smod not in ("__main__", "builtins"):
                    stype = smod + '.' + stype
        else:
            stype = str(self.etype)
        lines.append("%s%s\n" % (stype,
                                 value_str and ": %s" % value_str or ""))

        return lines

    def format_exception(self):
        lines = list(self.format_tb())
        if lines:
            lines.insert(0, "Traceback (most recent call last):\n")
        lines.extend(self.format_exception_only())
        return lines



from stacked import Stacked


class LoggingCompat(Stacked):

    def __init__(self, handler, tbcompat, **options):
        super(LoggingCompat, self).__init__()
        self.handler = handler
        self.tbcompat = tbcompat
        self.options = options

        formatter = self.handler.formatter

        # this is shit but we're stuck with the stdlib implementation since
        # it caches the result of formatException which we don't want to do
        # as it will screw up other formatters who are expecting a regular
        # traceback
        _format = formatter.format

        def format(record):
            record.exc_text = None
            formatted = _format(record)
            record.exc_text = None
            return formatted

        self._register_patch(formatter, "format", format)

        self._register_patch(formatter, "formatException",
                             self.formatException)

    def formatException(self, ei):
        return str(self.tbcompat._factory(*ei, **self.options))



#!/usr/bin/env python

import os

from setuptools import setup

# under test we do not want the nose entry point installed because it screws up
# coverage as the xtraceback module is imported too early
if "XTRACEBACK_NO_NOSE" in os.environ:
    entry_points = None
else:
    entry_points = {
        "nose.plugins": ("xtraceback=xtraceback.nosextraceback:NoseXTraceback",),
    }

README = open(os.path.join(os.path.dirname(__file__), "README.rst"))

setup(name="xtraceback",
      version="0.4.0-rc1",
      description="A verbose traceback formatter",
      long_description="\n".join(README.read()),
      license="MIT",
      keywords="traceback exception nose",
      author="Arthur Noel",
      author_email="arthur@0compute.net",
      url="https://github.com/0compute/xtraceback",
      packages=("xtraceback",),
      install_requires=("stacked == 0.1.2",),
      extras_require=dict(syntax=("pygments")),
      tests_require=("nose", "pygments", "yanc"),
      test_suite="nose.collector",
      entry_points=entry_points,
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Environment :: Console",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Operating System :: POSIX",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.5",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.2",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: Implementation :: CPython",
          "Programming Language :: Python :: Implementation :: Jython",
      ],
      )




class XTracebackOptions(object):
    """
    XTraceback options

    :ivar stream: A file-like object that is the default for print_* methods
    :type stream: file
    :ivar color: Flag to force color on or off - if None look to whether the
        `stream` is a tty
    :type color: bool
    :ivar print_width: How many columns wide to print the screen - if None and
        `stream` is a tty on Unix then fill the available width
    :ivar offset: A stack frame offset - defaults to 0
    :type offset: int
    :ivar limit: Stack frame limit - if None the entire stack is returned
    :context: Lines of context that are included in traceback entries
    :type context: int
    """

    # default options
    _options = dict(
        stream=None,
        color=None,
        print_width=None,
        offset=0,
        limit=None,
        context=5,
        globals_module_include=None,
        # python 3
        chain=True,
        )

    # default flags
    _flags = dict(
        show_args=True,
        show_locals=True,
        show_globals=False,
        qualify_methods=True,
        shorten_filenames=True,
        )

    def __init__(self, **options):
        # options
        for key in self._options:
            value = options.pop(key, None)
            if value is None:
                value = self._options[key]
            setattr(self, key, value)
        # flags
        for key in self._flags:
            value = options.pop(key, None)
            if value is None:
                value = self._flags[key]
            else:
                value = bool(value)
            setattr(self, key, value)
        # there should be no more options
        if options:
            raise TypeError("Unsupported options: %r" % options)



import os


def monkeypatch():

    if not hasattr(os.path, "relpath"):

        # taken from py27 posixpath

        try:

            import posixpath as pp

            def relpath(path, start=pp.curdir):
                """Return a relative version of a path"""

                if not path:
                    raise ValueError("no path specified")

                start_list = [x for x in pp.abspath(start).split(pp.sep) if x]
                path_list = [x for x in pp.abspath(path).split(pp.sep) if x]

                # Work out how much of the filepath is shared by start and path.
                i = len(pp.commonprefix([start_list, path_list]))

                rel_list = [pp.pardir] * (len(start_list) - i) + path_list[i:]
                if not rel_list:
                    return pp.curdir
                return pp.join(*rel_list)

            os.path.relpath = relpath

        except ImportError:
            pass



"""
XTraceback plugin for nose
"""

try:
    from nose.plugins import Plugin
except ImportError:
    pass
else:

    from .tracebackcompat import TracebackCompat

    class NoseXTraceback(Plugin):
        """Extended traceback for error output."""

        name = "xtraceback"

        # must be before capture otherwise we don't see the real sys.stdout
        score = 600

        _options = (
            ("globals",
             "Include globals in tracebacks [%s]",
             "store_true"),
            ("globals_include",
             "Include only globals in this namespace [%s]",
             "store"),
            ("color",
             "Show color tracebacks - one of on,off,auto (default=auto) [%s]",
             "store"),
        )

        def options(self, parser, env):
            super(NoseXTraceback, self).options(parser, env)
            for name, doc, action in self._options:
                env_opt = "NOSE_XTRACEBACK_%s" % name.upper()
                parser.add_option("--xtraceback-%s" % name.replace("_", "-"),
                                  action=action,
                                  dest="xtraceback_%s" % name,
                                  default=env.get(env_opt),
                                  help=doc % env_opt)

        def configure(self, options, conf):
            super(NoseXTraceback, self).configure(options, conf)
            for option in self._options:
                name = "xtraceback_%s" % option[0]
                setattr(self, name, getattr(options, name))
            if self.xtraceback_color is not None:
                if self.xtraceback_color == "on":
                    self.xtraceback_color = True
                else:
                    self.xtraceback_color = False

        def begin(self):
            options = dict(
                color=self.xtraceback_color,
                stream=self.conf.stream,
                show_globals=self.xtraceback_globals,
                globals_module_include=self.xtraceback_globals_include
            )
            self.compat = TracebackCompat(**options)
            self.compat.__enter__()

        def finalize(self, result):
            self.compat.__exit__(None, None, None)



try:
    from pygments.lexer import bygroups, include, using
    from pygments.lexers.agile import PythonLexer, PythonTracebackLexer
    from pygments.token import Text, Name, Number, Generic, String, Operator
except ImportError:  # pragma: no cover
    # this is for nose coverage which does a recursive import on the package
    pass
else:

    BASE_NAME = r"[a-zA-Z_][a-zA-Z0-9_]*"

    class XPythonLexer(PythonLexer):

        tokens = PythonLexer.tokens.copy()

        tokens["classname"] = [
            ("'?[a-zA-Z_][a-zA-Z0-9_.]*'?", Name.Class, "#pop")
        ]

        # Marker __repr__
        ref = r"(<ref offset)(=)(\-\d+)( ?)((?:name)?)(=?)((?:%s)?)(>?)" \
            % BASE_NAME
        tokens["root"].insert(0, (ref, bygroups(Name.Builtin, Name.Operator,
                                                Number, Text, Name.Builtin,
                                                Name.Operator, Name.Variable,
                                                Name.Builtin)))

    class PythonXTracebackLexer(PythonTracebackLexer):

        tokens = {
            "root": [
                include("entry"),
                include("exception"),
                (r"^.*\n", Generic.Error),
            ],
            "entry": [
                (r"^Traceback \(most recent call last\):\n",
                 Generic.Error,
                 "frame"),
                # file - path is colored differently if under working directory
                (r'^(  File )((?:"[./<][^"]+")?)((?:"[^"]+")?)'
                 r'(, line )(\d+)((?:, in )?)(.*)(\n)',
                 bygroups(Generic.Error, Name.Builtin, Operator.Word,
                          Generic.Error, Number, Generic.Error, Name.Function,
                          Text),
                 "frame"),
            ],
            "exception": [
                (r"^(AssertionError: )(.+\n)", bygroups(Generic.Error,
                                                        using(XPythonLexer))),
                (r"^(%s:?)(.+\n)" % BASE_NAME, bygroups(Generic.Error,
                                                        String)),
            ],
            "frame": [
                include("entry"),
                include("exception"),
                # line of python code
                (r"^((?:-+>)?)( +)(\d+)(.+\n)",
                 bygroups(Generic.Error, Text, Number, using(XPythonLexer))),
                # variable continuation
                (r"^([ ]+)('[^']+')(: )(.*)([,}]?\n)",
                 bygroups(
                     Text, String, Name.Operator, using(XPythonLexer), Text
                 )),
                # variable
                (r"^([ ]+)((?:g:)?)(\**%s)( = )(.+\n)" % BASE_NAME,
                 bygroups(Text, Name.Builtin, Name.Variable, Name.Operator,
                          using(XPythonLexer))),
                # plain python
                (r"^(    )(.+)(\n)",
                 bygroups(Text, using(XPythonLexer), Text)),
            ],

        }
